#!/usr/bin/env python
"""Tests for `t_nextgen` package."""


class TestTNextgen:
    """Smoke tests of the package."""

    def test_example(self) -> None:
        """Smoke test."""
        assert True



"""General utility functions."""

from pathlib import Path
import time
import re


class Timer:
    """Timer object."""

    def __init__(self, duration: int | float = 10):
        """Timer class initializing function."""
        self.duration = float(duration)
        self.start = time.perf_counter()

    def reset(self) -> None:
        """Resets the elapsed time."""
        self.start = time.perf_counter()

    def increment(self, increment: int = 0) -> None:
        """Increments duration."""
        self.duration += increment

    @property
    def expired(self) -> bool:
        """Checks if timer is expired."""
        return time.perf_counter() - self.start > self.duration

    @property
    def not_expired(self) -> bool:
        """Checks if timer is not expired."""
        return not self.expired

    @property
    def at(self) -> float:
        """Returns elapsed time."""
        return time.perf_counter() - self.start


def wait_until_file_downloads(
    directory_path: Path,
    file_identifier: str,
    is_regex: bool = False,
    wait_time: int = 30,
) -> str:
    """Wait until a file with the specified name or matching the regex pattern is found in the given directory."""
    if isinstance(directory_path, str):
        directory_path = Path(directory_path)
    timer = Timer(wait_time)
    while timer.not_expired:
        for file_path in directory_path.iterdir():
            if (is_regex and re.match(file_identifier, file_path.name)) or file_path.name == file_identifier:
                return str(file_path)
        time.sleep(1)
    raise AssertionError(
        f"{'File matching the pattern' if is_regex else 'File'} not found: {file_identifier} after {wait_time} seconds."
    )



"""COB Details model."""

from decimal import Decimal
from t_object import ThoughtfulObject
from t_desktop.config import IS_WINDOWS_OS

if IS_WINDOWS_OS:
    from pywinauto.controls.uia_controls import EditWrapper

from typing import Optional


class COBDetails(ThoughtfulObject):
    """Represents details of a COB entry."""

    rsn_code: str = ""
    rsn_amt: Decimal = Decimal(0)
    rsn_amt_element: Optional[EditWrapper] = None



"""PaymentEntry module."""
import contextlib
import ctypes
import _ctypes
from datetime import datetime
from logging import Logger
import os
from pathlib import Path
import re
from typing import Optional
from typing import Any, List
from retry import retry
from t_ocr.ocr import OCR

from decimal import Decimal
from t_desktop.config import IS_WINDOWS_OS
from t_desktop.utils.capture_screenshot import capture_screenshot

if IS_WINDOWS_OS:
    from pywinauto.application import WindowSpecification
    from pywinauto.controls.uia_controls import ListItemWrapper, EditWrapper
    from pywinauto.findwindows import ElementNotFoundError
    from pywinauto.base_wrapper import ElementNotEnabled
    from pywinauto.keyboard import send_keys
    from pywinauto.timings import wait_until, TimeoutError

from t_desktop.decorators import retry_if_pywin_error
from t_desktop.decorators import capture_screenshot_if_pywin_error
from t_ocr import Textract
from PIL import Image, ImageFilter

from t_nextgen.nextgen_window import NextGenWindow
from t_nextgen.utils.convert_to_decimal import convert_to_decimal
from t_nextgen.utils.opencv import OpenCV
from t_nextgen.exceptions import (
    AdjFieldNotFoundException,
    BalanceFieldNotFoundException,
    Ins1FieldNotFoundException,
    Ins2FieldNotFoundException,
    Ins3FieldNotFoundException,
    LnItemRsnsNotUpdatedException,
    PayFieldNotFoundException,
    ServiceLineMissingRequiredFieldError,
    TransactionNumberNotFoundError,
    DeductFieldNotFoundException,
    FieldNotFoundException,
    StatusNotUpdatedException,
    ButtonIsStillEnabledException,
    PayerNotUpdatedException,
)


class PaymentEntryWindow(NextGenWindow):
    """PaymentEntry class with methods to interact with payment entry window."""

    def __init__(self, app_path: str, logger: Logger, textract: Optional[Textract]):
        """Initialize the PaymentEntryWindow class.

        Args:
            app_path (str): The application path.
            textract (Optional[Textract]): The Textract object.
            logger (Logger): The logger object.

        """
        self.textract = textract
        super().__init__(app_path, logger)

    @property
    def window(self) -> WindowSpecification:
        """Return the PaymentEntry window element."""
        return self.desktop_app.dialog.child_window(title="PaymentEntry", control_type="Window")

    def set_all_service_lines_to_appeal(self) -> None:
        """This method sets all the service lines to appeal."""
        # Look at set_encounter_to_appeal_function
        self.logger.info("Setting Encounter to Appeal...")
        self.click_open_button()
        rows = self.get_service_rows()
        for row in rows:
            if self.get_pay_field(row):
                self.select_status_in_dropdown(status="Appeal", row=row, status_key="A")

    def get_pay_field(self, row: ListItemWrapper) -> str:
        """This method gets the pay field text in a specific row.

        Args:
            row (ListItemWrapper): list item wrapper with the row
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "pay" in window_text.lower():
                return item.get_value()
        raise PayFieldNotFoundException("Pay field not found in the row")

    def get_cpt4_field(self, row: ListItemWrapper) -> str:
        """This method gets the cpt 4 field text in a specific row."""
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "cpt4" in window_text.lower():
                element = row.descendants(title=window_text, control_type="Edit")[0]
                return element.get_value()
        raise FieldNotFoundException("cpt 4 field not found in the row")

    def get_qty_charge_field(self, row: ListItemWrapper) -> str:
        """This method gets the qty/charge text in a specific row."""
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "qty/charge" in window_text.lower():
                return item.get_value()
        raise FieldNotFoundException("qty/charge not found in the row")

    def get_date_field(self, row: ListItemWrapper) -> str:
        """This method gets the Date field text in a specific row."""
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "date" in window_text.lower():
                return item.get_value()
        raise FieldNotFoundException("Date field not found in the row")

    def is_line_voided(
        self, row: ListItemWrapper, column_name: str = "Rendering", image_path: Path = Path("C:\\Temp\\")
    ) -> bool:
        """Check if the line is voided.

        Args:
            row: The row to be checked
            column_name: The column name to be checked (optional - default: "Rendering")

        Returns:
            bool: True if the line is voided, False otherwise
        """
        image_path.mkdir(parents=True, exist_ok=True)
        element = row.descendants(title=column_name)[0]
        is_line_voided = self.is_text_strikethrough(element, image_path)
        if is_line_voided:
            self.logger.debug("The line is voided")
        return is_line_voided

    @staticmethod
    def is_text_strikethrough(element: Any, image_path: Path) -> bool:
        """Check if the text in the element is strikethrough.

        Args:
            element: The element to be checked
        Returns:
            bool: True if the text is strikethrough, False otherwise
            image_path: Path to save the image
        """
        # element name with the hour min and sec
        element_name_with_date = f"{element.window_text()}_{datetime.now().strftime('%H_%M_%S_%f')}.png"
        element_image_path = os.path.join(image_path, element_name_with_date)
        image_element = element.capture_as_image()
        # Convert the image to grayscale
        grayscale_image = image_element.convert("L")
        # Sharpen the grayscale image to enhance the edges of the text
        sharpened_image = grayscale_image.filter(ImageFilter.SHARPEN)
        # Convert the sharpened image back to RGB
        sharpened_rgb_image = sharpened_image.convert("RGB")
        sharpened_rgb_image.save(element_image_path)
        is_strikethrough_present = OpenCV.check_strikethrough_in_text_image(element_image_path)
        os.remove(element_image_path)
        return is_strikethrough_present

    @retry(ButtonIsStillEnabledException, tries=3, delay=1)
    def click_open_button(self) -> None:
        """This method clicks on the open button."""
        self.logger.debug("Clicking on the open button.")
        try:
            open_button = self.window.child_window(title="_cmdAction_1", control_type="Button")
            if open_button.is_enabled():
                with contextlib.suppress(_ctypes.COMError):
                    open_button.click_input()
                    open_button.wait_not("enabled", timeout=3, retry_interval=0.001)
        except TimeoutError:
            self.logger.error("Timeout error waiting for open button to be disabled.")
            raise ButtonIsStillEnabledException()

    def get_service_rows(self) -> list[ListItemWrapper]:
        """This method gets the service rows from the current payment entry window.

        Returns:
            rows (list): list of service rows
        """
        pane = self.window.child_window(title="lstPayLedger", control_type="Pane")
        rows = pane.wrapper_object().descendants(control_type="DataItem")
        self.logger.debug(f"Number of rows in Payment Entry: {len(rows)}")
        return rows

    @retry(LnItemRsnsNotUpdatedException, tries=3, delay=1)
    def insert_text_in_ln_item_rsns(self, text: str, row: ListItemWrapper) -> None:
        """This method inserts text in the Ln Item Rsns field.

        Args:
            text (str): text to be inserted. This text will have codes separated by commas (Ex:"PR2,CO45")
            row (ListItemWrapper): row to select. Will use the first row if not provided
        """
        self.logger.debug(f"Inserting text in Ln Item Rsns: {text}")
        if self.is_payer_a_patient():
            self.logger.info("The payer is Patient. Ln Item Rsns field is disabled")
            return

        self.remove_selected_items_from_ln_combobox(row)
        self.set_text_in_specific_row(row, "Ln Item Rsns", text, is_combobox=True)
        row = self._get_row_by_auto_id(row.automation_id())  # need to get the row with updated Ln Item Rsns
        next_gen_codes = row.descendants(title="Ln Item Rsns", control_type="Edit")[0].get_value()
        # next_gen_codes example :  'CO45  (Charges exceed fee arrangement); PR2  (Coinsurance amount)'
        codes_inserted = text.split(",")
        for code in codes_inserted:
            if code.strip() not in next_gen_codes:
                self.skipping_payer_aarp_pop_up_exist()
                raise LnItemRsnsNotUpdatedException(f"Error trying to insert {text} in Ln Item Rsns field.")
        return row

    @retry(LnItemRsnsNotUpdatedException, tries=3, delay=1)
    def add_codes_in_ln_item_rsns(self, text: str, row: ListItemWrapper) -> None:
        """Add one or more codes to Ln Item Rsns without removing existing ones.

        Args:
            text (str): Comma-separated codes to add (e.g., "PR2,CO45").
            row (ListItemWrapper): Row to update.
        """
        self.logger.debug(f"Adding codes to Ln Item Rsns: {text}")
        if self.is_payer_a_patient():
            self.logger.info("The payer is Patient. Ln Item Rsns field is disabled")
            return

        existing_codes = self._get_existing_ln_item_rsns_codes(row)
        new_codes = [c.strip().upper() for c in text.split(",") if c.strip()]
        combined_codes = existing_codes + [c for c in new_codes if c not in existing_codes]

        if combined_codes == existing_codes:
            self.logger.debug("Ln Item Rsns already contains requested codes; nothing to do.")
            return row

        combined_text = ",".join(combined_codes)

        self.remove_selected_items_from_ln_combobox(row)
        self.set_text_in_specific_row(row, "Ln Item Rsns", combined_text, is_combobox=True)

        row = self._get_row_by_auto_id(row.automation_id())
        next_gen_codes = (row.descendants(title="Ln Item Rsns", control_type="Edit")[0].get_value() or "").upper()

        for code in combined_codes:
            if code.strip().upper() not in next_gen_codes:
                self.skipping_payer_aarp_pop_up_exist()
                raise LnItemRsnsNotUpdatedException(f"Error trying to add {text} in Ln Item Rsns field.")

        return row

    def _get_existing_ln_item_rsns_codes(self, row: ListItemWrapper) -> list[str]:
        """Return the list of currently selected Ln Item Rsns codes for the given row.

        The NextGen control renders values like:
        'CO45  (Charges exceed fee arrangement); PR2  (Coinsurance amount)'.
        We parse and return only the raw codes: ['CO45', 'PR2'].
        """
        row = self.revalidate_service_row(row)
        value = row.descendants(title="Ln Item Rsns", control_type="Edit")[0].get_value()
        if value.strip() == "":
            return []

        parts = re.split(r"[;,]\s*", value)
        parts = [p.strip() for p in parts if p.strip()]

        codes: list[str] = []
        for part in parts:
            code_text = part.split("(")[0].strip()
            code_token = code_text.split()[0] if code_text else ""
            if code_token:
                codes.append(code_token.upper())

        return codes

    def remove_selected_items_from_status_combobox(self, row: ListItemWrapper) -> None:
        """Removes the selected items from the Status combobox.

        Args:
            row (ListItemWrapper): row to select
        """
        self.logger.debug("Removing selected items from Status combobox")
        row = self.revalidate_service_row(row)
        _input = row.descendants(title="Status", control_type="Edit")[0]

        if _input.get_value() == "":
            return
        with contextlib.suppress(_ctypes.COMError):
            _input.click_input(double=True)
        pane = self.window.child_window(title="_cboInput_0", control_type="Pane")
        combo_box = pane.child_window(control_type="ComboBox")
        combo_box.wait("ready", 2)
        self.desktop_app.dialog.child_window(
            auto_id="ColScrollRegion: 0, RowScrollRegion: 0", found_index=0
        ).right_click_input()
        send_keys("D")
        send_keys("{ESC}{TAB}")

    def is_payer_a_patient(self) -> bool:
        """This method checks if the payer is Patient.

        Returns:
            bool: True if payer is Patient, False otherwise
        """
        vse_container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        cbo_input = vse_container.child_window(auto_id="_cboInput_104", control_type="Pane")
        mcbo = cbo_input.child_window(auto_id="mcbo", control_type="ComboBox")
        payer = mcbo.child_window(control_type="Edit", found_index=0)
        return payer.get_value().lower() == "patient"

    @retry(StatusNotUpdatedException, tries=3, delay=1)
    def select_status_in_dropdown(
        self, status: str = "Appeal", row: ListItemWrapper | None = None, status_key: str = ""
    ) -> bool:
        """This method selects the specified status in the dropdown.

        Args:
            status (str): status to be selected (default is "Appeal")
            row (ListItemWrapper): row to select. Will use the first row if not provided
            status_key (str): Shortcut key for setting status (default sets to "Appeal" using "A"). Use
            only if first character of status is unique

        Returns:
            bool: True if the status was selected, False if it was already set
        """
        self.logger.debug(f"Selecting status in dropdown: {status}")
        self.click_open_button()
        row = self.revalidate_service_row(row)
        status_field = self.get_status_field(row)
        if status_field.get_value().lower() == status.lower():
            self.logger.debug(f"Status was already set. Value: {status}")
            return
        pane = self.open_status_dropdown_in_service_row(status_field, row)
        if len(status.split()) > 2:
            beginning, end = status.split()[0], status.split()[-1]
            status = self.get_status_in_dropdown_that_begin_with_and_end_with(beginning, end, pane)
        self.select_status_in_service_row(pane, status, status_key)
        row = self.revalidate_service_row(row)
        status_field = self.get_status_field(row)
        selected_status = status_field.get_value()
        if selected_status == "" or selected_status.lower() != status.lower():
            capture_screenshot(f"Failed_to_set_status_{status}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.png")
            raise StatusNotUpdatedException(f"Failed to select status {status} in dropdown")

    def get_statuses(self, pane: ListItemWrapper) -> list[str]:
        """This method gets all the statuses in the status dropdown.

        Args:
            pane (ListItemWrapper): Pane element containing the status dropdown.

        Returns:
            list[str]: A list of all the statuses in the dropdown.
        """
        combo_box = pane.child_window(control_type="ComboBox")
        combo_box_element = combo_box.wrapper_object()
        status_list = []
        for item in combo_box_element.descendants(control_type="ListItem"):
            status_list.append(item.window_text())
        return status_list

    def get_status_in_dropdown_that_begin_with_and_end_with(
        self, start_with: str, ends_with: str, pane: EditWrapper
    ) -> str:
        """This method gets the status in the drop down that begins with the value provided.

        Args:
            start_with (str): value to be searched for in the status column
            ends_with (str): value to be searched for in the status column
            pane (EditWrapper): Represents the pane element of the status field

        Returns:
            str: The selected status
        """
        self.logger.debug(f"Selecting status in dropdown that begins with: {start_with} and ends with {ends_with}")

        combo_box = pane.child_window(control_type="ComboBox")
        combo_box_element = combo_box.wrapper_object()
        for item in combo_box_element.descendants(control_type="ListItem"):
            status = item.window_text()
            if status.lower().startswith(start_with.lower()) and status.lower().endswith(ends_with.lower()):
                return status
        return ValueError(f"Could not find status that begins with: {start_with} and ends with {ends_with}")

    def select_status_in_service_row(self, pane: WindowSpecification, status: str, status_key: str = "") -> None:
        """Set status in the status column of current service row.

        Args:
            pane (WindowSpecification): Represents the pane element of the status field
            status (str): Status we want to set
            status_key (str, optional): Shortcut Key for setting status using first character of status. Use
            only if first character of status is unique
        """
        if status_key:
            send_keys(status_key)
            send_keys("{ENTER}")
        else:
            combo_box = pane.child_window(control_type="ComboBox")
            self.desktop_app.set_text(
                combo_box.child_window(auto_id="mcbo_EmbeddableTextBox", control_type="Edit"), status
            )
            with contextlib.suppress(_ctypes.COMError):
                pane.click_input()
                send_keys("+{TAB}")

    @retry(ElementNotFoundError, tries=3, delay=1)
    def open_status_dropdown_in_service_row(self, status_field: EditWrapper, row: ListItemWrapper) -> EditWrapper:
        """Open the status dropdown for the provided status field.

        Args:
            status_field (EditWrapper): the status field whose dropdown is to be opened
            row (ListItemWrapper): row to revalidate

        Returns:
            WindowSpecification: Represents the pane element of the status field
        """
        self.logger.debug("Opening status dropdown in service row")
        for i in range(3):
            row = self.revalidate_service_row(row)
            status_field = self.get_status_field(row)
            status_field.click_input()
        pane = self.desktop_app.dialog.child_window(title="_cboInput_0", control_type="Pane")
        return pane

    def get_status_field(self, row: ListItemWrapper) -> EditWrapper:
        """Get status field from service row.

        Args:
            row (ListItemWrapper): The service row from which to fetch status field

        Returns:
            EditWrapper: the status field
        """
        status_field = row.descendants(title="Status", control_type="Edit")[0]
        return status_field

    def select_item_in_line_dropdown(self, row: ListItemWrapper, column_name: str, item: str) -> None:
        """This method selects the specified item in the dropdown.

        Args:
            row (ListItemWrapper): row to use
            column_name (str): column name
            item (str): item to be selected
        """
        self.logger.debug(f"Selecting item in dropdown: {item}")
        row = self._get_row_by_auto_id(row.automation_id())
        input = row.descendants(title=column_name, control_type="Edit")[0]
        if input.get_value() == str(item):
            return
        self.desktop_app.mouse_click_element(input, "left")
        self.desktop_app.mouse_click_element(input, "left")
        pane = self.desktop_app.dialog.child_window(title="_cboInput_0", control_type="Pane")
        combo_box = pane.child_window(control_type="ComboBox")
        combo_box.wait("ready", 2)
        edit_text = combo_box.child_window(control_type="Edit", found_index=0)
        self.desktop_app.set_text(edit_text, item)
        edit_text.type_keys("{ENTER}")
        combo_box.wait("ready", 2)
        input.set_focus()  # focus again in input to remove the dropdown from screen

    def _get_row_by_auto_id(self, auto_id: str) -> ListItemWrapper:
        """This method gets the row by auto_id from the current payment entry window.

        Args:
            auto_id (str): auto_id of the row
        Returns:
            row (ListItemWrapper): row object
        """
        pane = self.desktop_app.dialog.child_window(title="lstPayLedger", control_type="Pane")
        row = pane.child_window(control_type="DataItem", auto_id=auto_id)
        return row.wrapper_object()

    def set_transaction_notes_text(self, notes: str) -> None:
        """This method sets the transaction notes text.

        Args:
            notes (str): transaction notes text
        """
        self.logger.debug(f"Setting transaction notes text: {notes}")
        transaction_note_input = self.window.child_window(auto_id="txtTransactionNotes", control_type="Edit")
        with contextlib.suppress(_ctypes.COMError):
            transaction_note_input.set_edit_text(notes)

    def click_recalc_button(self, click_recalc_button: bool = True) -> None:
        """Clicks the recalc button."""
        if click_recalc_button:
            self.logger.debug("Clicking recalc button in Payment Entry Window.")
            with contextlib.suppress(_ctypes.COMError):
                self.window.child_window(title="_cmdAction_6", control_type="Button").click()
        else:
            self.logger.warning("Skipping recalc button click to prevent changes.")

    def click_save_button(self, click_save_button: bool = True) -> None:
        """Clicks the save button."""
        if click_save_button:
            self.logger.debug("Clicking save button on Payment Entry Window.")
            with contextlib.suppress(_ctypes.COMError):
                self.window.child_window(title="_cmdAction_2", control_type="Button").click()
            self.verify_and_handle_modal("Please specify a status for each line item payments and adjustment.")
        else:
            self.logger.warning("Clicking cancel in payment entry to prevent changes.")
            self.click_cancel()

    def click_cancel(self) -> None:
        """Clicks the cancel button on the payment entry window."""
        self.logger.debug("Clicking cancel button on Payment Entry Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(title="_cmdAction_3", control_type="Button").click()

    def verify_and_handle_modal(self, expected_text: str) -> None:
        """Verifies a modal's text and raises an exception if it matches the expected text.

        Args:
            expected_text (str): The expected text to verify in the modal.
        """
        success, modal_text = self.desktop_app.close_modal(timeout=2)
        if not success:
            return
        self.logger.debug(f"Modal appeared. Text: {modal_text}")
        if modal_text.upper() == expected_text.upper():
            message = f"'{expected_text}' found in the modal. Raising exception."
            self.click_cancel()
            raise ServiceLineMissingRequiredFieldError(message)

    def set_text_in_specific_row_to_adj_field(self, row: ListItemWrapper, text: str | float) -> None:
        """This method sets the text in a specific row to the adj field.

        The adj column can have different names on it. All of them have pay in the text.

        Args:
            row (ListItemWrapper): list item wrapper with the row
            text (str): text to be inserted
        """
        self.logger.debug(f"Setting text in Adj field: {text}")
        items = row.children(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "adj" in window_text.lower():
                self.set_text_in_specific_row(row, window_text, text)
                self.skipping_payer_aarp_pop_up_exist()
                return
        raise AdjFieldNotFoundException("Adj field not found in the row")

    def set_text_in_specific_row(
        self, row: ListItemWrapper, element_name: str, text: str | float, is_combobox: bool = False
    ) -> None:
        """This method sets the text in a specific row.

        Args:
            row (ListItemWrapper): list item wrapper with the row
            element_name (str): element name
            text (str): text to be inserted
            is_combobox (bool): True if the element is a combobox, False otherwise
        """
        row = self.revalidate_service_row(row)
        element = row.descendants(title=element_name, control_type="Edit")[0]
        if element.get_value() == str(text):
            return
        self.desktop_app.mouse_double_click_element(element)
        with contextlib.suppress(_ctypes.COMError):
            wait_until(timeout=7, retry_interval=0.7, func=element.is_visible)
        self.desktop_app.mouse_double_click_element(element)
        with contextlib.suppress(_ctypes.COMError):
            wait_until(timeout=10, retry_interval=1, func=element.is_editable)
        self.desktop_app.set_text(element, text)
        self.desktop_app.mouse_double_click_element(element)

    def revalidate_service_row(self, row: ListItemWrapper, retries: int = 2) -> ListItemWrapper:
        """This method revalidates the service row in case the state of the row is not as expected.

        (i.e. the row is not visible or the coordinates are (0, 0))

        Args:
            row (ListItemWrapper): row to revalidate
            retries (int): number of retries
        Returns:
            row (ListItemWrapper): revalidated row
        """
        row_coords = self.desktop_app.get_element_coordinates(row)
        if row_coords == (0, 0) and retries > 0:
            refreshed_row = self._get_row_by_auto_id(row.automation_id())
            return self.revalidate_service_row(refreshed_row, retries=retries - 1)
        return row

    def get_claim_date(self) -> str:
        """This method get Claim Date Input value.

        Returns:
             str: claim input date value
        """
        claim_date_input = self.desktop_app.dialog.child_window(auto_id="mtxtDate", control_type="Edit")
        claim_date_value = claim_date_input.get_value()
        return claim_date_value

    def get_transaction_notes_text(self) -> str:
        """This method gets the transaction notes text.

        Returns:
            str: transaction notes text
        """
        transaction_note_input = self.window.child_window(auto_id="txtTransactionNotes", control_type="Edit")
        notes_value = transaction_note_input.get_value()
        return notes_value

    def get_pay_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """This method gets a value from a specific row.

        Args:
            row (ListItemWrapper): list item wrapper with the row
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "pay" in window_text.lower():
                element = row.descendants(title=window_text, control_type="Edit")[0]
                return element.get_value()
        raise PayFieldNotFoundException("Pay Field not found")

    def get_adj_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Adj" field in the row.

        Args
            matching_row (ListItemWrapper): if this argument is given, the method will
            search for the adj in this line.

        Returns:
            str: The value in the "Adj" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "adj" in window_text.lower():
                adj = row.children(title=window_text, control_type="Edit")[0]
                return adj.get_value()
        raise AdjFieldNotFoundException("Adj Field not found")

    def get_ins1_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Ins 1" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Ins 1 in this line.

        Returns:
            str: The value in the "Ins 1" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "ins 1" in window_text.lower():
                ins1 = row.children(title=window_text, control_type="Edit")[0]
                return ins1.get_value()
        raise Ins1FieldNotFoundException("Ins 1 field not found")

    def get_ins2_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Ins 2" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Ins 2 in this line.

        Returns:
            str: The value in the "Ins 2" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "ins 2" in window_text.lower():
                ins2 = row.children(title=window_text, control_type="Edit")[0]
                return ins2.get_value()
        raise Ins2FieldNotFoundException("Ins 2 field not found")

    def get_ins3_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Ins 3" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Ins 3 in this line.

        Returns:
            str: The value in the "Ins 3" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "ins 3" in window_text.lower():
                ins3 = row.children(title=window_text, control_type="Edit")[0]
                return ins3.get_value()
        raise Ins3FieldNotFoundException("Ins 3 field not found")

    def get_balance_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Balance" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Balance in this line.

        Returns:
            str: The value in the "Balance" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "balance" in window_text.lower():
                balance = row.children(title=window_text, control_type="Edit")[0]
                return balance.get_value()
        raise BalanceFieldNotFoundException("Balance field not found")

    @retry(ElementNotFoundError, tries=3, delay=0.5)
    def remove_selected_items_from_ln_combobox(self, row: ListItemWrapper) -> None:
        """Removes the selected items from the Ln Item Rsns combobox.

        Args:
            row (ListItemWrapper): row to select
        """
        row = self.revalidate_service_row(row)
        _input = row.descendants(title="Ln Item Rsns", control_type="Edit")[0]
        if _input.get_value() == "":
            return
        with contextlib.suppress(_ctypes.COMError):
            _input.double_click_input()
            _input.click_input()
            _input.click_input()
        with contextlib.suppress(_ctypes.COMError):
            self.desktop_app.app.window(title="MainForm").child_window(
                title="", control_type="Window"
            ).right_click_input()
        send_keys("D")
        send_keys("{ESC}{TAB}")

    def maximize_window(self) -> None:
        """Maximize the Payment Entry Window."""
        self.logger.debug("Maximizing Payment Entry Window.")
        self.window.maximize()

    @retry_if_pywin_error(retries=2, delay=5)
    @capture_screenshot_if_pywin_error()
    def click_new_button(self) -> None:
        """Click New Button."""
        self.logger.debug("Clicking New Button.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(title="_cmdAction_0", control_type="Button").click()
            self.desktop_app.wait_until_element_visible(
                title="_txtActiveSearch_103", control_type="Edit", timeout=30, retries=3
            )

    def enter_text_in_enc_clm_field(self, text: str) -> None:
        """This method inserts text in the Enc/Clm #.

        Args:
            text (str): text to be inserted
        """
        self.logger.debug(f"Inserting text in Enc/Clm #: {text}")
        with contextlib.suppress(_ctypes.COMError):
            container = self.window.child_window(auto_id="vseContainer", control_type="Text")
            self.desktop_app.set_text(container.child_window(title="_txtActiveSearch_103", control_type="Edit"), text)
            send_keys("{ENTER}")

    def enter_text_in_tracking_field(self, text: str) -> None:
        """This method inserts text in the Enc/Clm #.

        Args:
            text (str): text to be inserted
        """
        self.logger.debug(f"Inserting text in Tracking #: {text}")
        with contextlib.suppress(_ctypes.COMError):
            container = self.window.child_window(auto_id="vseContainer", control_type="Text")
            self.desktop_app.set_text(container.child_window(title="_txtDataEntry_105", control_type="Edit"), text)
            send_keys("{ENTER}")

    def click_close_nextgen_alerts(self, timeout: int = 5, retry_interval: float = 0.001) -> None:
        """Click close button on NextGen Alerts.

        Args:
            timeout (int, optional): Time to wait for alert to appear. Defaults to 5.
            retry_interval (float, optional): Time between retries when checking for alert. Defaults to 1.
        """
        alert = self.desktop_app.dialog.child_window(title="NextGen Alerts", control_type="Window")
        if alert.exists(timeout=timeout, retry_interval=retry_interval):
            self.logger.debug("Clicking close on NextGen Alerts")
            with contextlib.suppress(_ctypes.COMError):
                alert.child_window(title="Close", control_type="Button").click()

    def click_yes_encounter_in_history_status_pop_up(self, timeout: int = 1, window_title: str = "NextGen") -> None:
        """Click yes button in the encounter in history status pop-up.

        Args:
            timeout (int, optional): Time to wait for the pop-up to appear. Defaults to 1.
            window_title (str, optional): Title of the pop-up window. Defaults to "NextGen".
        """
        alert_window = self.desktop_app.dialog.child_window(title=window_title, control_type="Window")

        if alert_window.child_window(
            title="The encounter is in a history status.  Posting a transaction(s) against this encounter "
            "will reinstate the encounter to a billed status.  Do you want to continue?",
            control_type="Text",
        ).exists(timeout=timeout, retry_interval=0.001):
            self.logger.debug("Clicking Yes in Encounter in History Status pop-up")
            with contextlib.suppress(_ctypes.COMError):
                alert_window.child_window(title="Yes", control_type="Button").click()

    def _capture_pages_data_as_image_in_window(self) -> Image:
        """Capture Pages Data as Image.

        Returns:
            Image: the captured image
        """
        container = self.desktop_app.dialog.child_window(auto_id="vseContainer", control_type="Text")
        custom_control = container.child_window(auto_id="_lblDisplayField_0", control_type="Text")
        return custom_control.capture_as_image()

    def get_pay_column_header_text(self) -> str:
        """Get the text of the Pay column header.

        Returns:
            str: The Pay column header text.
        """
        pay_column_header = self.window.child_window(auto_id="[Column Header] Column18", control_type="HeaderItem")
        pay_column_header_text = pay_column_header.window_text()
        self.logger.debug(f"Pay column header text: {pay_column_header_text}")
        return pay_column_header_text

    def get_adj_column_header_text(self) -> str:
        """Get the text of the Adj column header.

        Returns:
            str: The Adj column header text.
        """
        adj_column_header = self.window.child_window(auto_id="[Column Header] Column19", control_type="HeaderItem")
        adj_column_header_text = adj_column_header.window_text()
        self.logger.debug(f"Adj column header text: {adj_column_header_text}")
        return adj_column_header_text

    def get_pages_data(self, screenshot_destination: Path, text_extraction_service: OCR) -> str:
        """Extract Pages Dataw.

        Returns:
            str: the pages data as text
            text_extraction_service (OCR): Object of OCR class
        """
        image = self._capture_pages_data_as_image_in_window()
        return self.get_data_from_image(image, screenshot_destination, text_extraction_service)

    def get_data_from_image(
        self, image: Image.Image, screenshot_destination: Path, text_extraction_service: OCR
    ) -> str:
        """Extract text from image after preprocessing it.

        Args:
            image (Image.Image): The image from which to extract text.
            parsing_settings (PSM): The parsing settings.
            text_extraction_service (OCR): Object of OCR class

        Returns:
            str: The extracted text.
        """
        screenshot_name = f'temp_screenshot_{datetime.now().strftime("%Y%m%d_%H%M%S")}.png'
        screenshot_location = screenshot_destination / screenshot_name
        image.save(screenshot_location)

        # Perform OCR
        ocrd_image = text_extraction_service.read_image_page(screenshot_location, cache_data=False)

        # Clean up the temporary image file
        screenshot_location.unlink()

        return ocrd_image.full_text

    def get_transaction_number(
        self, screenshot_destination: Path, text_extraction_services: List[OCR]
    ) -> Optional[str]:
        """Extracts and returns the transaction number using fallback approach.

        First tries to get the transaction number using handle method,
        then falls back to OCR if the handle method fails.

        Returns:
            Optional[str]: The transaction number if found, otherwise raises TransactionNumberNotFoundError.
        """
        # First approach: Try to get transaction number using handle method
        txn_handle = self.get_handle_of_transaction_number_field()
        txn_value = self.get_text_from_hwnd(txn_handle)
        match = re.search(r"(\d+)\s*of\s*\d+", txn_value)
        if match:
            txn_value = "00" + match.group(1)
            self.logger.debug(f"Transaction Number (handle method): {txn_value}")
            return txn_value

        # Second approach: Fallback to OCR method
        self.logger.debug("Handle method failed, trying OCR approach")
        self.maximize_window()
        self.desktop_app.wait_until_element_visible(auto_id="vseContainer", control_type="Text")
        for service in text_extraction_services:
            page_data = self.get_pages_data(screenshot_destination, service)
            match = re.search(r"(\d+)\s*of\s*\d+", page_data)
            if match:
                txn_value = "00" + match.group(1)
                self.logger.debug(f"Transaction Number (OCR method): {txn_value}")
                return txn_value

        raise TransactionNumberNotFoundError("Transaction number not found in the page data.")

    def get_handle_of_transaction_number_field(self) -> int:
        """This method gets the handle of transaction number field."""
        element = self.window.child_window(auto_id="_lblDisplayField_0", control_type="Text")
        return element.handle

    def get_text_from_hwnd(self, hwnd: int) -> str:
        """Get text from a window/control handle using SendMessage."""
        # Text size in control
        length = ctypes.windll.user32.SendMessageW(hwnd, 0x000E, 0, 0)
        if length == 0:
            return ""
        # Create buffer to receive the text
        buffer = ctypes.create_unicode_buffer(length + 1)
        # Fill the buffer with text
        ctypes.windll.user32.SendMessageW(hwnd, 0x000D, length + 1, buffer)
        return buffer.value

    def click_shield_icon(self) -> None:
        """Click shield icon in payment entry window."""
        self.logger.debug("Clicking shield icon in Payment Entry Window.")
        pnl_small_tool = self.window.child_window(title="pnlSmallTool", control_type="Text")
        with contextlib.suppress(_ctypes.COMError):
            pnl_small_tool.child_window(auto_id="_cmdSmallTool_03", control_type="Button").click_input()

    def click_balance_control_icon(self) -> None:
        """Click balance control icon in payment entry window."""
        self.logger.debug("Clicking balance control icon in Payment Entry Window.")
        pnl_small_tool = self.window.child_window(title="pnlSmallTool", control_type="Text")
        with contextlib.suppress(_ctypes.COMError):
            pnl_small_tool.child_window(auto_id="_cmdSmallTool_04", control_type="Button").click_input()

    @retry(PayerNotUpdatedException, tries=3, delay=1)
    def select_payer_in_drop_down(self, payer: str) -> None:
        """This method selects the payer in the drop down.

        Args:
            payer (str): payer to be selected
        """
        self.logger.debug(f"Selecting payer in dropdown: {payer}")
        container = self.window.child_window(auto_id="vseContainer", control_type="Text", top_level_only=True)
        pane = container.child_window(title="_cboInput_104", control_type="Pane", top_level_only=True)
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox", top_level_only=True)
        pane_object = pane.wrapper_object()
        options = pane_object.descendants(control_type="ListItem")
        for option in options:
            value = option.window_text()
            if payer.lower() in value.lower():
                self.desktop_app.set_text(
                    combo_box.child_window(control_type="Edit", top_level_only=True, found_index=0), value
                )
                break
        nextgen_payer = combo_box.child_window(auto_id="[Editor] Edit Area", control_type="Edit").get_value()
        if payer.lower() not in nextgen_payer.lower():
            self.logger.debug(f"Failed to select payer {payer} in dropdown. Retrying...")
            capture_screenshot(f"Failed_to_set_payer_{payer}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.png")
            raise PayerNotUpdatedException(f"Failed to select payer {payer} in dropdown")

    def get_payers_from_dropdown(self) -> list:
        """Get all the Payer element from the dropdown.

        Returns:
            list: List of payers as ListItem
        """
        container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        pane = container.child_window(title="_cboInput_104", control_type="Pane")
        pane_object = pane.wrapper_object()
        return pane_object.descendants(control_type="ListItem")

    def set_text_in_allowed(self, row: ListItemWrapper, allowed_amt: Decimal) -> None:
        """Update the Allowed field for a specific service line."""
        self.logger.debug(f"Setting Allowed value in specific row: {allowed_amt}")
        items = row.children(control_type="Edit")
        allowed_element_name = None
        for item in items:
            window_text = item.window_text()
            if "allowed" in window_text.lower():
                allowed_element_name = window_text
                self.set_text_in_specific_row(row, allowed_element_name, str(allowed_amt))
                self.skipping_payer_aarp_pop_up_exist()
                return

    def set_text_in_specific_row_to_pay_field(self, row: ListItemWrapper, text: str | float) -> None:
        """This method sets the text in a specific row to the pay field.

        Args:
            row (ListItemWrapper): list item wrapper with the row
            text (str): text to be inserted
        """
        self.logger.debug(f"Setting text in Pay field: {text}")
        with contextlib.suppress(_ctypes.COMError):
            str_text = str(convert_to_decimal(text))
            items = row.children(control_type="Edit")
            payment_element_name = None
            for item in items:
                window_text = item.window_text()
                if "pay" in window_text.lower():
                    payment_element_name = window_text
                    self.set_text_in_specific_row(row, payment_element_name, str_text)
                    self.skipping_payer_aarp_pop_up_exist()
                    return
            raise PayFieldNotFoundException("Pay field not found in the row")

    def skipping_payer_aarp_pop_up_exist(self, timeout: int = 1) -> bool:
        """This method checks if the skipping payer AARP pop up exists. If it does, it clicks on the ok button.

        Returns:
            bool: True if the pop up exists, False otherwise
        """
        alert_window = self.desktop_app.dialog.child_window(auto_id="65535", control_type="Text")
        if alert_window.exists(timeout=timeout, retry_interval=0.001):
            with contextlib.suppress(_ctypes.COMError):
                self.desktop_app.dialog.child_window(title="OK", control_type="Button").click()
            self.logger.debug("Clicked on OK button in Payer AARP pop up.")
            return True
        return False

    def get_next_gen_payer(self) -> str:
        """Get NextGen Payer on Payment Entry Window."""
        vse_container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        cbo_input = vse_container.child_window(auto_id="_cboInput_104", control_type="Pane")
        mcbo = cbo_input.child_window(auto_id="mcbo", control_type="ComboBox", found_index=0)
        nextgen_payer = mcbo.child_window(auto_id="[Editor] Edit Area", control_type="Edit").get_value()
        match = re.search(r"^(.*?)\/", nextgen_payer)
        if match:
            return match.group(1)
        self.logger.debug(f"Payer: {nextgen_payer}")
        return nextgen_payer

    def get_deduct_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Deduct" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Balance in this line.

        Returns:
            str: The value in the "Deduct" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "deduct" in window_text.lower():
                deduct = row.children(title=window_text, control_type="Edit")[0]
                self.logger.debug(f"Deduct value: {deduct.get_value()}")
                return deduct.get_value()
        raise DeductFieldNotFoundException("Deduct field not found")

    def set_text_in_deduct_field(self, row: ListItemWrapper, deduct_amt: Decimal) -> None:
        """Update the Deduct field for a specific service line."""
        self.logger.debug(f"Setting Deduct value in specific row: {deduct_amt}")
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "deduct" in window_text.lower():
                allowed_element_name = window_text
                self.set_text_in_specific_row(row, allowed_element_name, str(deduct_amt))
                self.skipping_payer_aarp_pop_up_exist()
                return

    def refresh_and_enter_encounter_num_and_trn(self, encounter_num: str, payment_trn: str) -> None:
        """This function closes and reopens the payment.

        Args:
            encounter_num (str): Encounter number
            payment_trn (str): Payment TRN
        """
        self.logger.debug("Refreshing the PaymentEntry window.")
        with contextlib.suppress(_ctypes.COMError):
            self.logger.debug("Clean the PaymentEntry window with ALT+L.")
            self.window.type_keys("%l")
            self.logger.debug(f"Entering encounter number: {encounter_num} in PaymentEntry window.")
            self.enter_text_in_enc_clm_field(encounter_num)
            self.click_close_nextgen_alerts()
            self.click_yes_encounter_in_history_status_pop_up()
            self.enter_text_in_tracking_field(payment_trn)
            self.logger.debug("Payment window cleaned and tracking number entered successfully.")

    def select_pay_code_in_drop_down(self, option_to_select: str) -> None:
        """This method selects the pay code in the drop down.

        Args:
            option_to_select (str): pay code to be selected
        """
        self.logger.debug(f"Selecting payer in dropdown: {option_to_select}")
        container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        pane = container.child_window(title="_cboInput_112", control_type="Pane")
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox")
        pane_object = pane.wrapper_object()
        options = pane_object.descendants(control_type="ListItem")
        for option in options:
            value = option.window_text()
            if option_to_select.lower() in value.lower():
                self.desktop_app.set_text(combo_box.child_window(control_type="Edit", found_index=0), value)
                return

    def select_adj_code_in_drop_down(self, option_to_select: str) -> None:
        """Select the adjustment code from the dropdown.

        Args:
            option_to_select (str): the adjustment code to be selected
        """
        container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        pane = container.child_window(title="_cboInput_113", control_type="Pane")
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox")
        pane_object = pane.wrapper_object()
        options = pane_object.descendants(control_type="ListItem")
        for option in options:
            value = option.window_text()
            if option_to_select.lower() in value.lower():
                self.desktop_app.set_text(combo_box.child_window(control_type="Edit", found_index=0), value)
                return

    def close_payment_entry_window(self) -> None:
        """Click close payment entry window."""
        title_bar = self.window.child_window(control_type="TitleBar")
        title_bar.child_window(title="Close", control_type="Button").click_input()

    def insert_control_number_in_resub(self, control_number: str) -> None:
        """This method inserts the control number in the resubmission field.

        Args:
            control_number (str): _description_
        """
        self.desktop_app.set_text(
            self.window.child_window(title="_txtDataEntry_116", control_type="Edit", top_level_only=True),
            control_number,
        )

    def click_show_all_lines_check_box(self) -> None:
        """This method clicks on the show all lines check box."""
        check_box = self.window.child_window(title="chkShowAll", control_type="CheckBox")
        self.desktop_app.toggle_checkbox(check_box)

    def check_bad_debt(self, timeout: int = 2, window_title: str = "NextGen") -> bool:
        """This method checks if the encounter is pre-listed for bad debt.

        Returns:
            bool: True if the pop up exists, False otherwise
        """
        alert_window = self.desktop_app.dialog.child_window(title=window_title, control_type="Window")

        bad_debt_text = alert_window.child_window(
            title="This encounter is pre-listed for bad debt or in bad debt.  You may not edit any information.",
            control_type="Text",
        )
        if bad_debt_text.exists(timeout=timeout, retry_interval=0.001):
            alert_window.child_window(title="OK", control_type="Button").click_input()
            return True
        return False

    def check_invalid_encounter_pop_up_exist(self, timeout: int = 2, window_title: str = "NextGen") -> bool:
        """This method checks if the invalid encounter pop up exists.

        Args:
            timeout (int, optional): Time to wait for the pop-up to appear. Defaults to 2.

        Returns:
            bool: True if the pop up exists, False otherwise
        """
        alert_window = self.desktop_app.dialog.child_window(title=window_title, control_type="Window")

        invalid_encounter_text_1 = alert_window.child_window(
            title="The Encounter number entered is invalid.", control_type="Text", top_level_only=True
        )
        if invalid_encounter_text_1.exists(timeout=timeout, retry_interval=0.001) or alert_window.child_window(
            title="Invalid encounter number!", control_type="Text"
        ).exists(timeout=1, retry_interval=0.001):
            alert_window.child_window(title="OK", control_type="Button").click()
            return True
        return False

    def get_tracking_date(self) -> str:
        """Get the value of the Date input under the Tracking input.

        Returns:
            str: The Date field value.
        """
        self.logger.debug("Getting Date field value under Tracking input.")
        container = self.desktop_app.dialog.child_window(auto_id="vseContainer", control_type="Text")
        date_input = container.child_window(auto_id="mtxtDate", control_type="Edit")
        date_value = date_input.get_value()
        self.logger.debug(f"Date field value found: {date_value}")
        return date_value

    def get_enc_clm_edit_value(self) -> str:
        """This method gets the value by element name in edit control type.

        Args:
            element_name (str): element name

        Returns:
            str: value
        """
        pane = self.desktop_app.dialog.child_window(auto_id="_cboInput_103", control_type="Pane", top_level_only=True)
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox")
        edit_control = combo_box.child_window(control_type="Edit", found_index=0)
        enc_clm = edit_control.get_value()
        return enc_clm

    def get_resub_value(self) -> str:
        """This method quickly gets the value from the resub field.

        Returns:
            str | None: resub value if found, otherwise None
        """
        edit = self.desktop_app.dialog.child_window(title="PaymentEntry", control_type="Window").child_window(
            auto_id="_txtDataEntry_116", control_type="Edit"
        )
        return str(edit.get_value())

    def payment_entry_window_opened_in_encounter(self, num: str, resub: str) -> bool:
        """Check if Payment Entry is opened for a given encounter number and resub value.

        Args:
            num (str): Encounter number to verify.
            resub (str): Expected resub value.

        Returns:
            bool: True if Payment Entry matches both encounter and resub, else False.
        """
        self.logger.debug(f"Checking if payment entry window is opened for encounter: {num}")
        try:
            enc_clm = self.get_enc_clm_edit_value()
        except (ElementNotFoundError, ElementNotEnabled, TimeoutError):
            self.logger.info("Payment Entry window not opened or encounter field not readable.")
            return False

        if not enc_clm or num not in enc_clm:
            return False

        resub_value = self.get_resub_value()
        if str(resub) != str(resub_value):
            self.logger.debug(f"Resub mismatch or not readable: expected={resub!r}, got={resub_value!r}")
            return False

        self.logger.debug("Payment Entry is opened for the expected encounter and resub.")
        return True

    def delete_allowed_field(self, row: WindowSpecification) -> None:
        """Delete (set to empty) the Allowed field of the given service line."""
        self.logger.debug("Deleting Allowed field (set to empty).")
        row = self.revalidate_service_row(row)
        items = row.children(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "allowed" in window_text.lower():
                element = row.descendants(title=window_text, control_type="Edit")[0]
                element.click_input()
                send_keys("^a{DELETE}")
                if element.get_value() != "":
                    element.click_input()
                    send_keys("^a{DELETE}")
                return
        raise FieldNotFoundException("Allowed field not found in the row")

    def delete_adj_field(self, row: WindowSpecification) -> None:
        """Delete (set to empty) the Adj field of the given service line."""
        self.logger.debug("Deleting Adj field (set to empty).")
        row = self.revalidate_service_row(row)
        items = row.children(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "adj" in window_text.lower():
                element = row.descendants(title=window_text, control_type="Edit")[0]
                element.click_input()
                send_keys("^a{DELETE}")
                if element.get_value() != "":
                    element.click_input()
                    send_keys("^a{DELETE}")
                return
        raise AdjFieldNotFoundException("Adj field not found in the row")

    def delete_pay_field(self, row: WindowSpecification) -> None:
        """Delete (set to empty) the Pay field of the given service line."""
        self.logger.debug("Deleting Pay field (set to empty).")
        row = self.revalidate_service_row(row)
        items = row.children(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "pay" in window_text.lower():
                element = row.descendants(title=window_text, control_type="Edit")[0]
                element.click_input()
                send_keys("^a{DELETE}")
                if element.get_value() != "":
                    element.click_input()
                    send_keys("^a{DELETE}")
                return
        raise PayFieldNotFoundException("Pay field not found in the row")



"""This module contains decorators for the NextGen application."""
import time
from functools import wraps
import _ctypes

from t_nextgen.utils.logger import logger


def click_ok_if_only_option() -> callable:
    """This function is a decorator that wraps the click_ok_if_only_option function.

    Returns:
        callable: The wrapped function.
    """

    def decorator(func: callable) -> callable:
        @wraps(func)
        def wrapper(self: object, *args, **kwargs) -> callable:
            if hasattr(self.desktop_app, "click_ok_if_only_option"):
                start_time = time.time()
                self.desktop_app.click_ok_if_only_option(max_calls=1, modal_exists_timeout=1, sleep=1)
                logger.debug(f"Time taken: {time.time() - start_time}")
            else:
                logger.warning("The 'click_ok_if_only_option' method is not defined.")
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


def reconnect_to_app_and_retry() -> callable:
    """This function is a decorator that wraps the reconnect to app."""

    def decorator(func: callable) -> callable:
        @wraps(func)
        def wrapper(self: object, *args, **kwargs) -> callable:
            try:
                return func(self, *args, **kwargs)
            except _ctypes.COMError as e:
                logger.warning(f"Error: {e}")
                logger.info("Reconnecting to the application...")
                time.sleep(1)
                self.desktop_app.app = self.desktop_app.connect_to_app(title="MainForm", timeout=20)
                logger.info("Reconnected to the application.")
                return func(self, *args, **kwargs)

        return wrapper

    return decorator



"""Transaction Ledger Module."""
from decimal import Decimal
from t_desktop.config import IS_WINDOWS_OS

from t_nextgen.utils.convert_to_decimal import convert_to_decimal

if IS_WINDOWS_OS:
    from pywinauto.controls.uia_controls import ListItemWrapper
    from pywinauto.application import WindowSpecification
import contextlib
from t_desktop.decorators import retry_if_pywin_error
import _ctypes
from t_nextgen.nextgen_window import NextGenWindow
from t_nextgen.exceptions import SearchEncounterNotFoundException, LockedBatchError, HeaderButtonNotFoundException


class TransactionLedgerWindow(NextGenWindow):
    """TransactionLedger class with methods to interact with transaction ledger window."""

    @property
    def window(self) -> WindowSpecification:
        """Return the BatchLedger window element."""
        return self.desktop_app.dialog.child_window(auto_id="BatchLedger", control_type="Window")

    def _get_amount(self, title: str, auto_id: str) -> str:
        """Fetch Billed/Allowed/Payment Amount from Window.

        Args:
            title (str): title of the element
            auto_id (str): auto_id of the element

        Returns:
            str: amount
        """
        amt = self.window.child_window(title=title, auto_id=auto_id, control_type="Edit").get_value()
        return amt.replace("$", "").replace(",", "").strip()

    def get_billed_amount(self) -> str:
        """Fetch Billed Amount.

        Returns:
            str: Billed Amount
        """
        return self._get_amount("Billed", "0")

    def get_allowed_amount(self) -> str:
        """Fetch Allowed Amount.

        Returns:
            str: allowed amount
        """
        return self._get_amount("Allowed", "1")

    def get_payment_amount(self) -> str:
        """Fetch Payment Amount.

        Returns:
            str: payment amount
        """
        return self._get_amount("Payment", "2")

    def get_adjustment_amount(self) -> str:
        """Fetch Adjustment Amount.

        Returns:
            str: adjustment amount
        """
        return self._get_amount("Adjustment", "3")

    @retry_if_pywin_error(retries=3, delay=2)  # type: ignore[misc]
    def click_header_from_menu_and_wait_for_batch_maintanance(self) -> None:
        """Clicks the Header button."""
        self.logger.debug("Clicking on the Header button.")
        header_button = self.window.child_window(title="cmdHeader", control_type="Button")
        if header_button.exists(timeout=2, retry_interval=0.001):
            with contextlib.suppress(_ctypes.COMError):
                self.desktop_app.mouse_click_element(header_button, "left")
        else:
            raise HeaderButtonNotFoundException("Header button not found")

        self.desktop_app.dialog.child_window(title="AMBatches", control_type="Window").wait("visible", 15)

    def filter_rows_by_source_value(self, rows: list[ListItemWrapper], value: str) -> list[dict]:
        """Filters rows based on the source value and ensures 'Billed' is not '0'."""
        filtered_rows = []
        for index, row in enumerate(rows):
            source_value = row.children(title="Source")[0].get_value()
            billed = row.children(title="Billed")[0].get_value()
            if self.is_row_valid(source_value, billed, value):
                filtered_rows.append({"element": row, "index": index})
        return filtered_rows

    def is_row_valid(self, source_value: str, billed: str, value: str) -> bool:
        """Validates whether a row matches the source value and is billed."""
        return source_value == value and billed != "0"

    def double_click_the_row_where_source_is(self, value: str) -> None:
        """This method double clicks on the row where the source is the value provided.

        Args:
            value (str): value to be searched for in the source column

        Returns:
            None
        """
        filtered_rows: list = []
        pane = self.window.child_window(title="lstLedger", control_type="Pane")
        rows = pane.wrapper_object().descendants(control_type="DataItem")
        filtered_rows = self.filter_rows_by_source_value(rows, value)
        if len(filtered_rows) > 0:
            self.logger.debug(f"Search encounter found for source: {value}")
            self.click_row(filtered_rows[0]["element"], filtered_rows[0]["index"], pane)
        else:
            self.logger.debug(f"Search encounter not found for source: {value}")
            raise SearchEncounterNotFoundException

    def click_row(self, row: ListItemWrapper, index: int, pane: WindowSpecification, offset: int = 3) -> None:
        """This method clicks on the transaction ledger row.

        Args:
            row (ListItemWrapper): row to be clicked
            index (int): index of the row
            pane (WindowSpecification): pane element where the row is located
            offset (int): offset to click on the row, default is 3

        Returns:
            None
        """
        self.logger.debug(f"Clicking on the row at index: {index}")
        pane.set_focus()
        page_size = 50 if self.window.is_maximized() else 20
        self.desktop_app.click_center_and_scroll(self.window, index + offset, page_size=page_size)
        row.double_click_input()

        try:
            self.desktop_app.dialog.child_window(auto_id="PaymentEntry", control_type="Window").wait(
                "visible", 5, retry_interval=0.001
            )
        except TimeoutError as e:
            modal = self.window.child_window(
                title="Security has been specified for this batch.  This batch of "
                "transactions is only accessible by its specified owner."
            )
            if modal.exists(timeout=1, retry_interval=0.001):
                raise LockedBatchError("Batch locked: Security has been specified for this batch!")
            raise e

    def get_row_elements(self) -> list[ListItemWrapper]:
        """Get all row elements.

        Returns:
            List[ListItemWrapper]: Represent the row elements
        """
        pane = self.window.child_window(auto_id="lstLedger", control_type="Pane")
        tree_view = pane.child_window(control_type="Tree")
        return tree_view.children(control_type="DataItem")

    def open_the_encounter(
        self,
        source_value: str,
        billed_amt: Decimal,
        tran_code_value: str,
        clm_status: str,
    ) -> None:
        """Double clicks on the row matching source, payment, and tran code.

        Args:
            source_value (str): The source value to match.
            billed_amt (Decimal): The Billed Amt to match.
            tran_code_value (str): The tran code value to match.
            clm_status (str | None): The CLM status to consider for filtering rows.

        Raises:
            SearchEncounterNotFoundException: If no matching row is found.
        """
        self.logger.debug(
            f"Opening the encounter. Source: {source_value}, Billed Amt: {billed_amt}, " f"tran_code: {tran_code_value}"
        )
        pane = self.window.child_window(auto_id="lstLedger", control_type="Pane")
        rows = pane.wrapper_object().descendants(control_type="DataItem")
        for index, row in enumerate(rows):
            try:
                row_source = row.children(title="Source")[0].get_value()
                row_billed = convert_to_decimal(row.children(title="Billed")[0].get_value())
                row_tran_code = row.children(title="Tran Code")[0].get_value()
            except (IndexError, AttributeError, ValueError) as e:
                self.logger.warning(f"Skipping row due to error: {e}")
                continue
            source_match = row_source == source_value
            billed_match = row_billed == billed_amt
            tran_code_match = tran_code_value in row_tran_code
            billed_match_clm_status = (clm_status == "22" and row_billed < 0) or (clm_status != "22" and row_billed > 0)

            if (source_match and billed_match and tran_code_match) and billed_match_clm_status:
                self.logger.debug(
                    f"Found matching row for source: {source_value}, "
                    f"Billed Amt: {billed_amt}, tran code: {tran_code_value}"
                )
                self.click_row(row, index, pane)
                return

        self.logger.debug(
            f"No matching row found for source: {source_value}, Billed Amt: {billed_amt}, tran code: {tran_code_value},"
            f" Clm Status: {clm_status}"
        )
        raise SearchEncounterNotFoundException



"""EncounterInsurance module."""
import contextlib
from t_desktop.config import IS_WINDOWS_OS

if IS_WINDOWS_OS:
    from pywinauto.application import WindowSpecification
    from pywinauto.controls.uia_controls import ListItemWrapper
    from pywinauto.timings import TimeoutError

from typing import Any
import _ctypes
from retry import retry

from t_nextgen.nextgen_window import NextGenWindow
from t_nextgen.ng_pm.windows.models.insurance_listing import InsuranceListing


class EncounterInsuranceWindow(NextGenWindow):
    """EncounterInsurance class with methods to interact with encounter insurance window."""

    @property
    def window(self) -> WindowSpecification:
        """Get Encounter Insurance Window."""
        return self.desktop_app.dialog.child_window(title="frmEncounterPayer", control_type="Window")

    def _get_tree_view_from_insurance_listing(self) -> WindowSpecification:
        """Get tree view of insurance listing from Encounter Insurance Selection Window.

        Returns:
            WindowSpecification: It contains the nested elements that have insurance listing data
        """
        pane = self.window.child_window(title="lstRelations", control_type="Pane")
        return pane.child_window(control_type="Tree")

    def get_insurance_listings_from_encounter_insurance_selection_window(self) -> list[InsuranceListing]:
        """Build Insurance Listing object from insurance listings present in encounter insurance selection window.

        Returns:
            List[InsuranceListing]: A list of InsuranceListing objects
        """
        tree_view = self._get_tree_view_from_insurance_listing()
        data_items = tree_view.children(control_type="DataItem")
        insurance_listings = []
        for item in data_items:
            data_grids = item.children(control_type="DataGrid")
            for grid in data_grids:
                payers = grid.children(control_type="DataItem")
                for curr_payer in payers:
                    curr_insurance_listing = InsuranceListing()
                    curr_cols = curr_payer.children(control_type="Edit")
                    for col in curr_cols:
                        if col.window_text() == "Column1":
                            payer = col.get_value()
                            curr_insurance_listing.payer_row = col
                            curr_insurance_listing.payer = payer
                        elif col.window_text() == "Column3":
                            policy_number = col.get_value()
                            curr_insurance_listing.policy_number = policy_number
                    insurance_listings.append(curr_insurance_listing)
        self.logger.debug(f"Insurance List: {insurance_listings}")
        return insurance_listings

    def select_payer_row_in_insurance_listing(self, payer_row: ListItemWrapper) -> None:
        """Click the payer row from insurance listing in encounter insurance selection window.

        Args:
            payer_row (ListItemWrapper): The payer row to be clicked
        """
        self.logger.debug("Selecting payer row in insurance listing")
        payer_row.type_keys("{DOWN}")
        with contextlib.suppress(_ctypes.COMError):
            payer_row.click_input()

    def _get_tree_view_from_selected_insurance(self) -> WindowSpecification:
        """Get tree view of selected insurance from Encounter Insurance Selection Window.

        Returns:
            WindowSpecification: It contains the nested elements that have selected insurance data
        """
        pane = self.window.child_window(title="lstPayers", control_type="Pane")
        return pane.child_window(control_type="Tree")

    def get_payers_from_selected_insurance(self) -> list[str]:
        """Get Payers from selected insurance in encounter insurance selection window.

        Returns:
            List[str]: List of Payer Names
        """
        tree_view = self._get_tree_view_from_selected_insurance()
        data_items = tree_view.children(control_type="DataItem")
        payers = []
        for item in data_items:
            payers.append(item.window_text())
        self.logger.debug(f"Payers from selected insurance: {payers}")
        return payers

    def click_right_arrow_in_encounter_insurance_selection_window(self) -> None:
        """Click right arrow in Encounter Insurance Selection Window."""
        with contextlib.suppress(_ctypes.COMError):
            self.desktop_app.dialog.child_window(title="cmdAdd", control_type="Button").click_input()

    def click_ok_in_encounter_insurance_selection_window(self) -> None:
        """Click ok in Encounter Insurance Selection Window."""
        self.logger.debug("Clicking ok in Encounter Insurance Selection Window")
        with contextlib.suppress(_ctypes.COMError):
            self.desktop_app.dialog.child_window(title="cmdOk", control_type="Button").click_input()

    def click_ok_encounter_insurance_selection_window_pop_up(self) -> None:
        """Click ok in Encounter Insurance Selection Window pop-up."""
        self.logger.debug("Clicking ok in Encounter Insurance Selection Window pop-up")
        if self.desktop_app.dialog.child_window(title="OK", control_type="Button").exists(timeout=2):
            with contextlib.suppress(_ctypes.COMError):
                self.desktop_app.dialog.child_window(title="OK", control_type="Button").click_input()

    def get_insurance_maintenance_window(self) -> WindowSpecification:
        """Get Insurance Maintenance Window.

        Returns:
            WindowSpecification: The Insurance Maintenance Window
        """
        insurance_window = self.desktop_app.dialog.child_window(title="frmAMIns", control_type="Window")
        if insurance_window.exists(timeout=2):
            return insurance_window

    def enter_co_pay(self, insurance_window: WindowSpecification, co_pay_amount: int) -> None:
        """Enter co-pay information in Insurance Maintenance Window.

        Args:
            insurance_window (WindowSpecification): The Insurance Maintenance Window
            co_pay_amount (int): The co-pay amount to be set
        """
        self.logger.debug(f"Entering co-pay amount: {co_pay_amount}")
        self.desktop_app.set_text(
            insurance_window.child_window(title="txtEncounterCopayAmt", control_type="Edit"), co_pay_amount
        )

    def click_ok(self, insurance_window: WindowSpecification) -> None:
        """Click ok in Insurance Maintenance Window.

        Args:
            insurance_window (WindowSpecification): The insurance maintenance window
        """
        self.logger.debug("Clicking OK in Insurance Maintenance Window")
        with contextlib.suppress(_ctypes.COMError):
            insurance_window.child_window(title="cmdOK", control_type="Button").click_input()

    @retry(tries=3, delay=3)
    def click_close_nextgen_alerts(self, title: str = "Edits") -> None:
        """Click close button on NextGen Alerts.

        Args:
            title (str): The title of the NextGen Alerts window. Defaults to "Edits".
        """
        self.logger.debug("Clicking close on NextGen Alerts")
        alert = self.desktop_app.dialog.child_window(title=title, control_type="Window")
        if alert.exists(timeout=2, retry_interval=0.001):
            with contextlib.suppress(_ctypes.COMError):
                alert.child_window(title="Close", control_type="Button").click()
                alert.wait_not("visible", timeout=5)

    def attach_unattached_payer(self, listing: InsuranceListing) -> None:
        """Handles the process of attaching an unattached payer to the insurance listing.

        Args:
            listing (InsuranceListing): The insurance listing containing the payer information.
        """
        self.logger.debug("Attaching unattached payer")
        self.select_payer_row_in_insurance_listing(listing.payer_row)
        self.click_right_arrow_in_encounter_insurance_selection_window()
        self.click_ok_in_encounter_insurance_selection_window()
        self.click_ok_encounter_insurance_selection_window_pop_up()

        insurance_window = self.get_insurance_maintenance_window()
        if insurance_window:
            self.logger.info("Insurance maintenance window opened")
            insurance_window.set_focus()
            self.enter_co_pay(insurance_window, 0)
            self.click_ok(insurance_window)
            self.click_close_nextgen_alerts()

        self.click_close_nextgen_alerts_encounter_insurance_selection_window()

    def click_close_nextgen_alerts_encounter_insurance_selection_window(self) -> None:
        """Click close button on NextGen Alerts."""
        self.logger.debug("Clicking close on NextGen Encounter Insurance Selection Alerts")
        alert = self.desktop_app.dialog.child_window(
            title="NextGen Encounter Insurance Selection Alerts", control_type="Window"
        )
        if alert.exists(timeout=2, retry_interval=0.001):
            alert.set_focus()
            with contextlib.suppress(_ctypes.COMError):
                alert.child_window(title="Close", control_type="Button").click()

    def get_payer_list(self) -> list[Any]:
        """Retrieves the payer list from the Encounter Insurance window."""
        payers = self.window.child_window(auto_id="lstRelations", control_type="Pane").child_window(
            title="Patient", control_type="DataItem"
        )

        options = payers.descendants(control_type="DataItem")
        self.logger.debug(f"Payers options list: {options}")
        return options

    @retry(exceptions=TimeoutError, tries=3, delay=1)
    def click_cancel_insurance_window(self) -> None:
        """Clicks the Cancel button in the Encounter Insurance window."""
        self.logger.debug("Clicking Cancel in Encounter Insurance window")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="cmdCancel", control_type="Button").click_input()
            self.window.wait_not("visible", timeout=5, retry_interval=0.001)



"""NgAppManager module."""

import contextlib
from logging import Logger
import time
from abc import ABCMeta
from typing import Optional, Any
import pyautogui
import _ctypes
from retry import retry
from t_desktop import DesktopApp
from t_desktop.config import IS_WINDOWS_OS
from t_desktop.decorators import retry_if_pywin_error

if IS_WINDOWS_OS:
    from pywinauto.timings import TimeoutError
    from pywinauto.controls.uia_controls import ComboBoxWrapper
    from pywinauto.application import Application, WindowSpecification
    from pywinauto.findwindows import ElementNotFoundError
    from pywinauto.controls.hwndwrapper import InvalidWindowHandle
    from pywinauto.controls.uia_controls import ListItemWrapper
    from pywinauto import mouse


from t_nextgen.utils.logger import logger as lib_logger


class ScopedSingletonMeta(ABCMeta):
    """SingletonMeta class."""

    _instances: dict = {}

    def __call__(cls, *args, **kwargs):
        """Ensures that only one instance of the class is created per `app_path`.

        This method overrides the default `__call__` method to control the instantiation
        of the class. When called, it checks if an instance of the class already exists
        for the specified `app_path` in `cls._instances`. If an instance does not exist
        for the provided `app_path`, it creates one and stores it in the `cls._instances`
        dictionary. Otherwise, it returns the existing instance associated with the
        `app_path`.

        The `app_path` can be provided as a keyword argument (`app_path`) or as the first
        positional argument. If no `app_path` is specified, a TypeError is raised.

        Args:
            *args: Positional arguments for the class instantiation.
            **kwargs: Keyword arguments for the class instantiation. Includes `app_path`.

        Returns:
            object: The single instance of the class associated with the given `app_path`.
        """
        app_path = ""
        if "app_path" in kwargs:
            app_path = kwargs["app_path"]
        elif len(args) > 0:
            app_path = args[0]
        else:
            raise TypeError("app_path argument is required.")
        if app_path not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[app_path] = instance
        return cls._instances[app_path]


class NGAppManager(DesktopApp, metaclass=ScopedSingletonMeta):
    """NgAppManager class."""

    def __init__(self, app_path: str, logger: Optional[Logger] = None):
        """Initialize the NgAppManager class.

        Args:
            app_path (str): The path to the NextGen application.
            logger (Optional[Logger], optional): The logger object to use for logging. Defaults to None.
        """
        self.logger = logger if logger is not None else lib_logger
        super().__init__(app_path)

    def close_modal(
        self,
        modal_title: str = "NextGen",
        buttons_to_try: list[str] = ["OK"],
        timeout: int = 5,
        retry_interval: int = 0.001,
    ) -> tuple[bool, str]:
        """Close a modal in the app by clicking on the given button and returning the modal text.

        Args:
            modal_title (str): Title of the modal.
            buttons_to_try (list[str]): List of button texts to try clicking, in order.
              Defaults to ["OK"] if the list is empty.
            timeout (int): Maximum amount of time to check for modal.
            retry_interval (int): The control is checked for existence this number of seconds.

        """
        modal = self.dialog.child_window(title=modal_title, control_type="Window", top_level_only=True)
        if modal.exists(timeout, retry_interval):
            modal_text_control = modal.child_window(control_type="Text", top_level_only=True)
            modal_text = modal_text_control.window_text()

            # Find and click the button
            buttons_list = modal.wrapper_object().descendants(control_type="Button")
            for button_to_click in buttons_to_try:
                for button in buttons_list:
                    button_text = button.texts()[0].upper()
                    if button_text == button_to_click.upper():
                        with contextlib.suppress(_ctypes.COMError):
                            button.click()
                        return True, modal_text
            self.logger.debug(f"No matching button found in modal for options {buttons_to_try}.")
            return False, modal_text
        else:
            return False, ""

    @retry(tries=3, delay=4)
    def connect_to_logged_in_screen(self) -> None:
        """Connect to the Next Gen App after it has been logged into."""
        try:
            return self.connect_to_app(title="MainForm", timeout=20)
        except Exception as e:
            self.click_on_next_gen_window()
            raise e

    @retry(tries=3, delay=3)
    def close_login_window(self, app_path: str) -> None:
        """Close the login window.

        Args:
            app_path (str): The path to the NextGen application.
        """
        app = Application(backend="uia").connect(path=app_path, timeout=5)
        for window in app.windows():
            if window.element_info.control_type == "Window":
                self.logger.debug(f"Closing Login window: {window.window_text()}")
                window.close()

    @retry_if_pywin_error()
    def open_file_menu(self) -> None:
        """This function opens the file menu."""
        file_menu = self.dialog.child_window(title="File", control_type="MenuItem", top_level_only=True, found_index=0)
        file_menu.click_input()
        self.dialog.child_window(title="FileDropDown").wait("visible", 30)

    @retry((TimeoutError, OSError), 3, 1)
    def connect_to_main_form(self) -> None:
        """Connect to nextgen main form.

        Raises:
            TimeoutError: Failed to connect to NextGen main form.
        """
        with contextlib.suppress(TimeoutError):
            self.app = self.connect_to_app(title="MainForm", timeout=20)
            return
        with contextlib.suppress(TimeoutError):
            self.app = self.connect_to_app(path=self.app_path, timeout=20)
            return
        raise TimeoutError("Timeout, failed to connect to NextGen main form.")

    @retry_if_pywin_error()
    def clear_out_windows_function(self, practice: str = "") -> None:
        """This function clears out windows."""
        child_windows = self.dialog.descendants(control_type="Window")

        self.logger.info(f"Found {len(child_windows)} child windows to close.")

        closed_windows_count = 0
        for child_window in child_windows:
            try:
                child_window.close()
                self.deal_with_unhandled_exception_popup(f"NextGen - {practice}")
                self.click_no_button_in_next_gen_alert()
                closed_windows_count += 1
            except (_ctypes.COMError, RuntimeError, AttributeError) as e:
                self.logger.error(f"Error closing window: {str(e)}")

        self.logger.debug(f"Closed {closed_windows_count} child windows.")

    def click_ok_if_only_option(
        self, call_count: int = 0, max_calls: int = 3, modal_exists_timeout: int = 3, sleep: float = 5
    ) -> bool:
        """Clicks the "OK" button on a modal if it is the only option and checks for additional modals.

        This function searches for a modal dialog with the title "NextGen". If found, it looks for "OK",
        "CLOSE", and "CANCEL" buttons. If "OK" is the only available button, it clicks it. The function
        will recursively check for additional modals up to a maximum of `max_calls`.

        Args:
            call_count (int, optional): _description_. Defaults to 0.
            max_calls (int, optional): _description_. Defaults to 3.
            modal_exists_timeout (int, optional): _description_. Defaults to 3.
            sleep (float, optional): _description_. Defaults to 5.
        """
        if call_count >= max_calls:
            return True
        try:
            modal = self.dialog.child_window(title="NextGen", control_type="Window")
            if modal.exists(timeout=modal_exists_timeout, retry_interval=0.001):
                wrapper_object = modal.wrapper_object()
                buttons_list = wrapper_object.descendants(control_type="Button")

                ok_button = None
                close_button = None
                more_buttons = False
                for button in buttons_list:
                    button_text = button.texts()[0].upper()
                    if button_text == "OK":
                        ok_button = button
                    elif button_text in ["CLOSE", "CANCEL"]:
                        close_button = button
                    else:
                        more_buttons = True
                if more_buttons:
                    return False
                if ok_button and close_button:
                    self.logger.debug("Ok button found. Clicking on it.")
                    texts = wrapper_object.descendants(control_type="Text")
                    if texts:
                        self.logger.debug(f"Text in the modal: {texts[0].texts()}")
                    with contextlib.suppress(_ctypes.COMError):
                        ok_button.click()
                    time.sleep(sleep)
                    self.click_ok_if_only_option(
                        call_count=call_count + 1, max_calls=max_calls
                    )  # Recursive call with incremented call_count
            else:
                return False
        except _ctypes.COMError:
            return True
        except (TimeoutError, ElementNotFoundError, IndexError):
            return False

    def click_ok_popup_getting_message(
        self, call_count: int = 0, max_calls: int = 3, modal_exists_timeout: int = 3, sleep: float = 5
    ) -> str:
        """Clicks the "OK" button on a modal if it is the only option and checks for additional modals.

        This function searches for a modal dialog with the title "NextGen". If found, it looks for "OK",
        "CLOSE", and "CANCEL" buttons. If "OK" is the only available button, it clicks it. The function
        will recursively check for additional modals up to a maximum of `max_calls`. And return the message
        in the modal.

        Args:
            call_count (int, optional): _description_. Defaults to 0.
            max_calls (int, optional): _description_. Defaults to 3.
            modal_exists_timeout (int, optional): _description_. Defaults to 3.
            sleep (float, optional): _description_. Defaults to 5.

        Returns:
            str: The text message from the modal dialog, or an empty string if modal was not found.
        """
        if call_count >= max_calls:
            return ""
        try:
            modal = self.dialog.child_window(title="NextGen", control_type="Window")
            if modal.exists(timeout=modal_exists_timeout, retry_interval=0.001):
                wrapper_object = modal.wrapper_object()
                buttons_list = wrapper_object.descendants(control_type="Button")

                ok_button = None
                close_button = None
                more_buttons = False
                for button in buttons_list:
                    button_text = button.texts()[0].upper()
                    if button_text == "OK":
                        ok_button = button
                    elif button_text in ["CLOSE", "CANCEL"]:
                        close_button = button
                    else:
                        more_buttons = True
                if more_buttons:
                    return ""
                if ok_button and close_button:
                    self.logger.debug("Ok button found. Clicking on it.")
                    texts = wrapper_object.descendants(control_type="Text")
                    modal_text = None
                    if texts:
                        modal_text = texts[0].texts()
                        self.logger.debug(f"Text in the modal: {modal_text}")
                    with contextlib.suppress(_ctypes.COMError):
                        ok_button.click()
                    time.sleep(sleep)
                    self.click_ok_popup_getting_message(
                        call_count=call_count + 1, max_calls=max_calls
                    )  # Recursive call with incremented call_count
                    return modal_text[0] if modal_text else ""
            else:
                return ""
        except _ctypes.COMError:
            return ""
        except (TimeoutError, ElementNotFoundError, IndexError):
            return ""

    def click_no_button_in_next_gen_alert(self) -> None:
        """Clicks the "No" button on the NextGen alert window if it exists."""
        no_button = self.dialog.child_window(title="NextGen", control_type="Window").child_window(
            auto_id="7", control_type="Button"
        )
        if no_button.exists(timeout=1, retry_interval=0.001):
            with contextlib.suppress(_ctypes.COMError):
                no_button.click()
            no_button.wait_not("visible", timeout=5)

    def select_value_from_combobox(
        self, combobox: ComboBoxWrapper, value: str, use_type_keys: bool = False, click_combobox: bool = True
    ) -> None:
        """Selects a value from a ComboBox by simulating a click and searching for the matching option.

        Args:
            combobox (ComboBoxWrapper): The ComboBox control in which to search for the value.
            value (str): The value to search for and select within the ComboBox.
            use_type_keys (bool, optional): If True, uses type_keys instead of set_text. Defaults to False.
            click_combobox (bool, optional): If True, clicks the combobox before selecting the value. Defaults to True.
        """
        if click_combobox:
            with contextlib.suppress(_ctypes.COMError):
                combobox.click_input()
        # Search for the matching option in the ComboBox
        option_to_select = self.get_option_to_select_from_combobox(combobox, value)
        # Select the option if found
        self.select_option_in_combobox(combobox, use_type_keys, option_to_select)

    def click_on_next_gen_window(self) -> None:
        """Click on the Next Gen window."""
        next_gen_app_coordinates = self.get_next_gen_app_coordinates()
        if next_gen_app_coordinates != (0, 0):
            self.mouse_click_element(self.app.top_window())

    def get_next_gen_app_coordinates(self) -> tuple:
        """Get the coordinates of the Next Gen App window.

        Returns:
            tuple: element coordinates
        """
        return self.get_element_coordinates(self.app.top_window())

    def double_click_row(self, row_elements: list[ListItemWrapper], row_index: int) -> None:
        """Double click a row.

        Args:
            row_elements (List[ListItemWrapper]): The list of row elements
            row_index (int): The row to be clicked. Follows zero based indexing.
        """
        with contextlib.suppress(_ctypes.COMError):
            row_elements[row_index].click_input(double=True)

    def click_down_n_times(self, n: int) -> None:
        """This method click down key 'n' times.

        Args:
            n (int): number of times to click
        """
        pyautogui.press("home")
        confidence_value = 3  # This is an extra number to ensure the row will be full visible, so let's scroll more
        self.logger.debug(f"Clicking down key {n + confidence_value} times.")
        for _ in range(n + confidence_value):
            pyautogui.press("down")

    def click_center_and_scroll(
        self, element: WindowSpecification, index: int, page_size: int = 50, scroll_step: int = 3
    ) -> None:
        """Clicks at the center of the given element and performs a downward scroll.

        Args:
            element: The element on which the click and scroll action is performed.
            index (int): The index to determine the scroll amount.
            page_size (int, optional): The size of the page to calculate scrolling. Defaults to 50
            scroll_step (int, optional): The step size for each scroll action. Defaults to 3.

        Raises:
            Exception: If any error occurs during the click or scroll operation,
        """
        rect = element.rectangle()
        center_x = rect.left + (rect.width() // 2)
        center_y = rect.top + (rect.height() // 2)
        mouse.click("left", (center_x, center_y))
        pyautogui.press("home")
        scroll_count = max(0, (index - page_size + scroll_step) // scroll_step)
        self.logger.debug(f"Index to scroll: {index}")
        self.logger.debug(f"Scroll count {scroll_count}")
        mouse.scroll(coords=(center_x, center_y), wheel_dist=-scroll_count)
        time.sleep(0.5)

    def click_button_by_element_name(self, element_name: str) -> None:
        """This method clicks on a button by element name.

        Args:
            element_name (str): element name
        """
        with contextlib.suppress(_ctypes.COMError):
            button = self.dialog.child_window(title=element_name, control_type="Button")
            self.mouse_click_element(button)

    def click_file_menu_button(self) -> None:
        """Click on the File menu button."""
        with contextlib.suppress(_ctypes.COMError):
            self.dialog.child_window(title="File", control_type="MenuItem").wait("exists", timeout=10)
            self.dialog.child_window(title="File", control_type="MenuItem").invoke()

    def get_file_menu(self) -> WindowSpecification:
        """Get file menu window control.

        Returns:
            WindowSpecification: File Menu control
        """
        self.dialog.child_window(title="File", control_type="MenuItem").wait("exists", timeout=10)
        return self.dialog.child_window(title="File", control_type="MenuItem")

    def click_processes_button_in_file_menu(self, file_menu: WindowSpecification) -> None:
        """Click on the Processes option in the File menu.

        Args:
            file_menu (WindowSpecification): File Menu control
        """
        with contextlib.suppress(_ctypes.COMError):
            file_menu.child_window(title="Processes", control_type="MenuItem", found_index=0).wait("exists", timeout=10)
            file_menu.child_window(title="Processes", control_type="MenuItem", found_index=0).invoke()

    def get_processes_menu_from_file_menu(self, file_menu: WindowSpecification) -> WindowSpecification:
        """Get the Processes Menu Control.

        Args:
            file_menu (WindowSpecification): File Menu control

        Returns:
            WindowSpecification: Processes Menu Control
        """
        file_menu.child_window(title="Processes", control_type="MenuItem", found_index=0).wait("exists", timeout=10)
        return file_menu.child_window(title="Processes", control_type="MenuItem", found_index=0)

    def close_digital_image_function(self) -> None:
        """This function clears out Digital Image windows."""
        windows_menu = self.dialog.child_window(title="Window", control_type="MenuItem")
        windows_menu.set_focus()
        windows_menu.select()

        child_windows = self.dialog.descendants(control_type="Window")

        closed_windows_count = 0
        for active_child_window in child_windows:
            window_name = active_child_window.window_text()
            if window_name.startswith("Digital Image"):
                if self.close_window_if_exist(window_name):
                    closed_windows_count += 1
                    return

    def select_option_in_combobox(self, combobox: ComboBoxWrapper, use_type_keys: bool, option_to_select: str) -> None:
        """This method selects an option in a ComboBox.

        Args:
            combobox (ComboBoxWrapper): ComboBox control.
            use_type_keys (bool): Boolean to determine if type_keys should be used.
            option_to_select (str): Option to select in the ComboBox.
        """
        self.logger.debug(f"Selecting option '{option_to_select}' in ComboBox.")
        if use_type_keys:
            combobox.type_keys(option_to_select)
            time.sleep(1)
            combobox.type_keys("{ENTER}")
        else:
            with contextlib.suppress(_ctypes.COMError):
                combobox.child_window(control_type="Edit", found_index=0).set_text(option_to_select)

    def get_option_to_select_from_combobox(self, combobox: ComboBoxWrapper, value: str) -> str:
        """Gets the option to select from the ComboBox.

        Args:
            combobox (ComboBoxWrapper): The ComboBox control in which to search for the value.
            value (str): The value to search for and select within the ComboBox.

        Returns:
            str: The option to select from the ComboBox.
        """
        option_to_select: str = ""
        for option in combobox.descendants(control_type="ListItem"):
            option_text = option.window_text().lower().strip()
            value = value.lower().strip()
            if value in option_text:
                option_to_select = option.window_text()
                break
        return option_to_select

    def click_on_combobox(self, combobox: ComboBoxWrapper) -> None:
        """Clicks the combobox.

        Args:
            combobox (ComboBoxWrapper): The ComboBox control to click.
        """
        with contextlib.suppress(_ctypes.COMError):
            combobox.click_input()

    def deal_with_unhandled_exception_popup(self, windows_title: str = "Microsoft .NET Framework") -> bool:
        """This method deals with the unhandled exception popup.

        Args:
            windows_title (str): The title of the window to check for.

        Returns:
            bool: True if the popup was found and handled, False otherwise.
        """
        with contextlib.suppress(ElementNotFoundError, InvalidWindowHandle):
            popup_app = Application(backend="win32").connect(title=windows_title)
            top_window = popup_app.top_window()
            if top_window.exists(timeout=1, retry_interval=0.001):
                top_window.child_window(title="&Continue").click()
                return True
            return False

    def get_element_coordinates(self, element: WindowSpecification) -> tuple:
        """This method gets the element coordinates.

        Args:
            element (ListItemWrapper): element

        Returns:
            tuple: element coordinates
        """
        return element.rectangle().mid_point()

    def safe_wait(self, kwargs: Any) -> bool | TimeoutError:
        """Waits for an element specified by `kwargs` to become visible in the `next_gen` dialog window.

        This function will attempt to wait for an element to become visible up to a maximum of 3 tries.
        If the element does not become visible within these tries, a `TimeoutError` will be raised.

        Parameters:
        kwargs : dict
            A dictionary of keyword arguments that specify the properties of the child window element
            to wait for. These arguments are passed to the `child_window` method.

        Returns:
        bool
            Returns True if the element becomes visible within the allowed tries.

        Raises:
        TimeoutError
            If the element does not become visible after 3 tries.

        Example:
        >>> kwargs = {"title": "Ok", "control_type": "Button"}
        >>> safe_wait(kwargs)
        True

        """
        max_tries = 3
        current_try = 0
        while not current_try > max_tries:
            try:
                self.dialog.child_window(**kwargs).wait("visible", 5)
                return True
            except (RuntimeError, TimeoutError):
                self.logger.debug("Element not yet loaded, waiting more 2 seconds")
                time.sleep(2)
                current_try += 1
        raise TimeoutError("Element was not loaded after 3 tries")

    def close_window_if_exist(self, window_title: str) -> bool:
        """Closes the specified window by its title.

        Args:
            window_title (str): The title of the window to close.
        """
        try:
            window = self.dialog.child_window(title=window_title, control_type="Window")
            window.close()
            self.click_no_button_in_next_gen_alert()
            self.logger.debug(f"Closed window: {window_title}")
            return True
        except ElementNotFoundError:
            self.logger.debug(f"Window '{window_title}' not found. It might already be closed.")
            return False

    def set_text(self, field: ListItemWrapper, text: str, max_retries: int = 4) -> None:
        """Set a text in a specific field and double check if the action was performed corretly, retry if not.

        Args:
            field (ListItemWrapper): Field where the text shoud be set.
            text (str): Text that will be set.
            max_retries (int, optional): Number of max retries attempt. Defaults to 4.
        """
        retries = 0
        while retries < max_retries:
            try:
                field.set_text(text)
                return
            except _ctypes.COMError:
                time.sleep(0.5)
                if field.get_value() == text:
                    return
            retries += 1
        self.logger.error(f"Failed to set text after {retries} retries")

    def toggle_checkbox(self, checkbox: ListItemWrapper, max_retries: int = 4) -> None:
        """Toggle a checkbox and double check if the action was performed corretly, retry if not.

        Args:
            checkbox (ListItemWrapper): Checkbox element that should be toggled.
            max_retries (int, optional): Number of max retries attempt. Defaults to 4.
        """
        retries = 0
        while retries < max_retries:
            try:
                if checkbox.get_toggle_state() == 1:
                    return
                checkbox.toggle()
                return
            except _ctypes.COMError:
                time.sleep(1)
                if checkbox.get_toggle_state() == 1:
                    return
            retries += 1
        self.logger.error(f"Failed to toggle checkbox after {retries} retries")

    def untoggle_checkbox(self, checkbox: ListItemWrapper, max_retries: int = 4) -> None:
        """Untoggle a checkbox and double check if the action was performed corretly, retry if not.

        Args:
            checkbox (ListItemWrapper): Checkbox element that should be untoggled.
            max_retries (int, optional): Number of max retries attempt. Defaults to 4.
        """
        retries = 0
        while retries < max_retries:
            try:
                if checkbox.get_toggle_state() == 0:
                    return
                checkbox.toggle()
                return
            except _ctypes.COMError:
                time.sleep(1)
                if checkbox.get_toggle_state() == 0:
                    return
            retries += 1
        self.logger.error(f"Failed to untoggle checkbox after {retries} retries")
